"""Paginators for iterating over API results.

Version Added:
    6.0:
    This replaces the paginator code in the old
    :py:mod:`reviewboard.hostingsvcs.utils.paginator` module.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Callable, Generic, TYPE_CHECKING, TypeVar, Union

from housekeeping import deprecate_non_keyword_only_args
from typing_extensions import NotRequired, TypeAlias, TypedDict

from reviewboard.deprecation import RemovedInReviewBoard10_0Warning

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typelets.funcs import KwargsDict

    from reviewboard.hostingsvcs.base import HostingServiceClient
    from reviewboard.hostingsvcs.base.http import HTTPHeaders, QueryArgs


_PageDataItemT = TypeVar('_PageDataItemT')
_PageDataT = TypeVar('_PageDataT')


#: Type alias for a normalize function for ProxyPaginator.
#:
#: Version Added:
#:     6.0
_ProxyNormalizePageDataFunc: TypeAlias = Callable[[Any],
                                                  Union[_PageDataT, None]]


class APIPaginatorPageData(TypedDict):
    """Data that can be returned from an APIPaginator.

    Version Added:
        6.0
    """

    #: The data from the page (generally as a list).
    #:
    #: Type:
    #:     object
    data: NotRequired[Any]

    #: The HTTP headers from the page response.
    #:
    #: Type:
    #:     dict
    headers: NotRequired[HTTPHeaders]

    #: The optional URL to the next page.
    #:
    #: Type:
    #:     str
    next_url: NotRequired[str | None]

    #: The optional limit on the number of items fetched on each page.
    #:
    #: Type:
    #:     int
    per_page: NotRequired[int | None]

    #: The optional URL to the previous page.
    #:
    #: Type:
    #:     str
    prev_url: NotRequired[str | None]

    #: The API response data.
    #:
    #: Type:
    #:     object
    response: NotRequired[object]

    #: The optional total number of items across all pages.
    #:
    #: Type:
    #:     int
    total_count: NotRequired[int | None]


class InvalidPageError(Exception):
    """An error representing an invalid page access.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.utils.paginator` to
          :py:mod:`reviewboard.hostingsvcs.base.paginator`.
    """


class BasePaginator(Generic[_PageDataItemT, _PageDataT]):
    """Base class for a paginator used in the hosting services code.

    This provides the basic state and stubbed functions for a simple
    paginator. Subclasses can build upon this to offer more advanced
    functionality.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.utils.paginator` to
          :py:mod:`reviewboard.hostingsvcs.base.paginator`.

        * This is now a generic, supporting typing for page data and items.
    """

    ######################
    # Instance variables #
    ######################

    #: The data for the current page.
    #:
    #: This is implementation-dependent, but will usually be a list. It must
    #: operate as a sequence of some kind.
    #:
    #: Type:
    #:     object
    page_data: _PageDataT | None

    #: The number of items to fetch per page.
    #:
    #: Type:
    #:     int
    per_page: int | None

    #: Keyword arguments to pass when making HTTP requests.
    #:
    #: Type:
    #:     dict
    request_kwargs: KwargsDict

    #: The starting page.
    #:
    #: Whether this is 0-based or 1-based depends on the hosting service.
    #:
    #: Type:
    #:     int
    start: int | None

    #: The total number of results across all pages.
    #:
    #: This will be ``None`` if the value isn't known.
    #:
    #: Type:
    #:     int
    total_count: int | None

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def __init__(
        self,
        *,
        start: (int | None) = None,
        per_page: (int | None) = None,
        request_kwargs: (KwargsDict | None) = None,
    ) -> None:
        """Initialize the paginator.

        Version Changed:
            8.0:
            Made arguments keyword-only.

        Args:
            start (int, optional):
                The starting page.

                Whether this is 0-based or 1-based depends on the hosting
                service.

            per_page (int, optional):
                The number of items per page.

            request_kwargs (dict, optional):
                Keyword arguments to pass when making a request.
        """
        self.start = start
        self.per_page = per_page
        self.page_data = None
        self.total_count = None
        self.request_kwargs = request_kwargs or {}

    @property
    def has_prev(self) -> bool:
        """Whether there's a previous page available.

        Subclasses must override this to provide a meaningful value.

        Type:
            bool
        """
        raise NotImplementedError

    @property
    def has_next(self) -> bool:
        """Whether there's a next page available.

        Subclasses must override this to provide a meaningful value.

        Type:
            bool
        """
        raise NotImplementedError

    def prev(self) -> _PageDataT | None:
        """Fetch the previous page, returning the page data.

        Subclasses must override this to provide the logic for fetching pages.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.

        Raises:
            InvalidPageError:
                There was no previous page to fetch.
        """
        raise NotImplementedError

    def next(self) -> _PageDataT | None:
        """Fetch the next page, returning the page data.

        Subclasses must override this to provide the logic for fetching pages.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.

        Raises:
            InvalidPageError:
                There was no next page to fetch.
        """
        raise NotImplementedError

    def iter_items(
        self,
        max_pages: (int | None) = None,
    ) -> Iterator[_PageDataItemT]:
        """Iterate through all items across pages.

        This will repeatedly fetch pages, iterating through all items and
        providing them to the caller.

        The maximum number of pages can be capped, to limit the impact on
        the server.

        Args:
            max_pages (int, optional):
                The maximum number of pages to iterate through.

        Yields:
            object:
            Each item from each page's payload.
        """
        for page in self.iter_pages(max_pages=max_pages):
            if page:
                assert isinstance(page, Sequence), (
                    f"page_data is not a sequence (it's a {type(page)}). This "
                    f"is either an unexpected result, or "
                    f"{type(self).__name__}.iter_items() needs to be "
                    "overridden."
                )

                yield from page

    def iter_pages(
        self,
        max_pages: (int | None) = None,
    ) -> Iterator[_PageDataT | None]:
        """Iterate through pages of results.

        This will repeatedly fetch pages, providing each parsed page payload
        to the caller.

        The maximum number of pages can be capped, to limit the impact on
        the server.

        Args:
            max_pages (int, optional):
                The maximum number of pages to iterate through.

        Yields:
            object:
            The parsed payload for each page.
        """
        try:
            if max_pages is None:
                while True:
                    yield self.page_data
                    self.next()
            else:
                for i in range(max_pages):
                    if i > 0:
                        self.next()

                    yield self.page_data
        except InvalidPageError:
            pass

    def __iter__(self) -> Iterator[_PageDataT | None]:
        """Iterate through pages of results.

        This is a simple wrapper for :py:meth:`iter_pages`.

        Yields:
            object:
            The parsed payload for each page.
        """
        yield from self.iter_pages()


class APIPaginator(BasePaginator[_PageDataItemT, _PageDataT]):
    """Handles pagination for API requests to a hosting service.

    Hosting services may provide subclasses of ``APIPaginator`` that can handle
    paginating their specific APIs. These make it easy to fetch pages of data
    from the API, and also works as a bridge for Review Board's web API
    resources.

    All ``APIPaginators`` are expected to take an instance of a
    :py:class:`~reviewboard.hostingsvcs.base.client.HostingServiceClient`
    subclass, and the starting URL (without any arguments for pagination).

    Subclasses can access the
    :py:class:`~reviewboard.hostingsvcs.base.client.HostingServiceClient`
    through the :py:attr:`client` member of the paginator in order to perform
    requests against the hosting service.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.utils.paginator` to
          :py:mod:`reviewboard.hostingsvcs.base.paginator`.

        * This is now a generic, supporting typing for page data and items.
    """

    #: Query parameter name for the start page in a request.
    #:
    #: This is optional. Clients can specify this to provide this as part
    #: of pagination queries.
    #:
    #: Type:
    #:     str
    start_query_param: (str | None) = None

    #: Query parameter name for the requested number of results per page.
    #:
    #: This is optional. Clients can specify this to provide this as part
    #: of pagination queries.
    #:
    #: Type:
    #:     str
    per_page_query_param: (str | None) = None

    ######################
    # Instance variables #
    ######################

    #: The hosting service client used to make requests.
    #:
    #: Type:
    #:     reviewboard.hostingsvcs.base.client.HostingServiceClient
    client: HostingServiceClient

    #: The URL for the next set of results in the page.
    #:
    #: Type:
    #:     str
    next_url: str | None

    #: HTTP headers returned for the current page.
    #:
    #: Type:
    #:     dict
    page_headers: HTTPHeaders | None

    #: The URL for the previous set of results in the page.
    #:
    #: Type:
    #:     str
    prev_url: str | None

    #: The URL used to fetch the current page of data.
    #:
    #: Type:
    #:     str
    url: str | None

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def __init__(
        self,
        *,
        client: HostingServiceClient | None,
        url: str,
        query_params: QueryArgs = {},
        start: (int | None) = None,
        per_page: (int | None) = None,
        request_kwargs: (KwargsDict | None) = None,
    ) -> None:
        """Initialize the paginator.

        Once initialized, the first page will be fetched automatically.

        Version Changed:
            8.0:
            * Made arguments keyword-only.
            * Explicitly listed out arguments that are passed to the base
              class.

        Args:
            client (reviewboard.hostingsvcs.base.client.HostingServiceClient):
                The hosting service client used to make requests.

            url (str):
                The URL used to make requests.

            query_params (dict):
                The query parameters to append to the URL for requests.

                This will be updated with :py:attr:`start_query_param`
                and :py:attr:`per_page_query_param`, if set.

            *args (tuple):
                Positional arguments for the parent constructor.

            start (int, optional):
                The starting page.

                Whether this is 0-based or 1-based depends on the hosting
                service.

            per_page (int, optional):
                The number of items per page.

            request_kwargs (dict, optional):
                Keyword arguments to pass when making a request.
        """
        super().__init__(
            start=start,
            per_page=per_page,
            request_kwargs=request_kwargs)

        self.client = client
        self.url = url
        self.prev_url = None
        self.next_url = None
        self.page_headers = None

        # Augment the URL with the provided query parameters.
        query_params = query_params.copy()

        if self.start_query_param and self.start:
            query_params[self.start_query_param] = self.start

        if self.per_page_query_param and self.per_page:
            query_params[self.per_page_query_param] = self.per_page

        self.request_kwargs.setdefault('query', {}).update(query_params)

        self._fetch_page()

    @property
    def has_prev(self) -> bool:
        """Whether there's a previous page available.

        Type:
            bool
        """
        return self.prev_url is not None

    @property
    def has_next(self) -> bool:
        """Whether there's a next page available.

        Type:
            bool
        """
        return self.next_url is not None

    def prev(self) -> _PageDataT | None:
        """Fetch the previous page, returning the page data.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.

        Raises:
            InvalidPageError:
                There was no previous page to fetch.
        """
        if not self.has_prev:
            raise InvalidPageError

        self.url = self.prev_url
        return self._fetch_page()

    def next(self) -> _PageDataT | None:
        """Fetch the next page, returning the page data.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.

        Raises:
            InvalidPageError:
                There was no next page to fetch.
        """
        if not self.has_next:
            raise InvalidPageError

        self.url = self.next_url
        return self._fetch_page()

    def fetch_url(
        self,
        url: str,
    ) -> APIPaginatorPageData:
        """Fetch the URL, returning information on the page.

        This must be implemented by subclasses.

        Args:
            url (str):
                The URL to fetch.

        Returns:
            dict:
            The pagination information with the above fields.

            See :py:class:`APIPaginatorPageData` for the supported fields.
        """
        raise NotImplementedError

    def _fetch_page(self) -> _PageDataT | None:
        """Fetch a page and extracts the information from it.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.
        """
        assert self.url is not None
        page_info = self.fetch_url(self.url)

        self.prev_url = page_info.get('prev_url')
        self.next_url = page_info.get('next_url')
        self.per_page = page_info.get('per_page', self.per_page)
        self.page_data = page_info.get('data')
        self.page_headers = page_info.get('headers', {})
        self.total_count = page_info.get('total_count')

        # Make sure the implementation sent the correct data to us.
        assert self.prev_url is None or isinstance(self.prev_url, str), (
            f'"prev_url" result from fetch_url() must be None or Unicode '
            f'string, not {type(self.prev_url)!r}'
        )

        assert self.next_url is None or isinstance(self.next_url, str), (
            f'"next_url" result from fetch_url() must be None or Unicode '
            f'string, not {type(self.next_url)!r}'
        )

        assert self.total_count is None or isinstance(self.total_count, int), (
            f'"total_count" result from fetch_url() must be None or int, not '
            f'{type(self.total_count)!r}'
        )

        assert self.per_page is None or isinstance(self.per_page, int), (
            f'"per_page" result from fetch_url() must be an int, not '
            f'{type(self.per_page)!r}'
        )

        assert isinstance(self.page_headers, dict), (
            f'"page_headers" result from fetch_url() must be a dictionary, '
            f'not {type(self.page_headers)!r}'
        )

        return self.page_data


class ProxyPaginator(BasePaginator[_PageDataItemT, _PageDataT]):
    """A paginator that proxies to another paginator, transforming data.

    This attaches to another paginator, forwarding all requests and proxying
    all data.

    ``ProxyPaginator`` can take the data returned from the other paginator and
    normalize it, transforming it into a new form.

    This is useful when a
    :py:class:`~reviewboard.hostingsvcs.base.hosting_service.
    BaseHostingService` wants to return a paginator to callers that represents
    data in a structured way, using an :py:class:`APIPaginator`'s raw payloads
    as a backing.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.utils.paginator` to
          :py:mod:`reviewboard.hostingsvcs.base.paginator`.

        * This is now a generic, supporting typing for page data and items.
    """

    ######################
    # Instance variables #
    ######################

    #: The paginator that this is a proxy for.
    #:
    #: Type:
    #:     BasePaginator
    paginator: BasePaginator[Any, Any]

    #: A function used to normalize a page of results from the paginator.
    #:
    #: Type:
    #:     callable
    normalize_page_data_func: (
        _ProxyNormalizePageDataFunc[_PageDataItemT] | None
    )

    def __init__(
        self,
        paginator: BasePaginator[Any, Any],
        normalize_page_data_func: (
            _ProxyNormalizePageDataFunc[_PageDataItemT] | None
        ) = None,
    ) -> None:
        """Initialize the paginator.

        Args:
            paginator (BasePaginator):
                The paginator that this is a proxy for.

            normalize_page_data_func (callable, optional):
                A function used to normalize a page of results from the
                paginator.
        """
        # NOTE: We're not calling BasePaginator here, because we're actually
        #       overriding all the properties it would set that we care about.
        self.paginator = paginator
        self.normalize_page_data_func = normalize_page_data_func
        self.page_data = self.normalize_page_data(self.paginator.page_data)

    @property
    def has_prev(self) -> bool:
        """Whether there's a previous page available.

        Type:
            bool
        """
        return self.paginator.has_prev

    @property
    def has_next(self) -> bool:
        """Whether there's a next page available.

        Type:
            bool
        """
        return self.paginator.has_next

    @property
    def per_page(self) -> int | None:
        """The number of items requested per page.

        Type:
            int
        """
        return self.paginator.per_page

    @property
    def total_count(self) -> int | None:
        """The number of items across all pages, if known.

        Type:
            int
        """
        return self.paginator.total_count

    def prev(self) -> _PageDataT | None:
        """Fetch the previous page, returning the page data.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.

        Raises:
            InvalidPageError:
                There was no previous page to fetch.
        """
        return self._process_page(self.paginator.prev())

    def next(self) -> _PageDataT | None:
        """Fetch the next page, returning the page data.

        Returns:
            object:
            The resulting page data.

            This will usually be a :py:class:`list`, but is
            implementation-dependent.

        Raises:
            InvalidPageError:
                There was no next page to fetch.
        """
        return self._process_page(self.paginator.next())

    def normalize_page_data(
        self,
        data: Any,
    ) -> _PageDataT | None:
        """Normalize a page of data.

        If :py:attr:`normalize_page_data_func` was passed on construction, this
        will call it, passing in the page data. That will then be returned.

        This can be overridden by subclasses that want to do more complex
        processing without requiring ``normalize_page_data_func`` to be
        passed in.

        Args:
            data (object):
                The data to normalize.

        Returns:
            object:
            The resulting data.
        """
        if callable(self.normalize_page_data_func):
            data = self.normalize_page_data_func(data)

        return data

    def _process_page(
        self,
        page_data: Any,
    ) -> _PageDataT | None:
        """Process a page of data.

        This will normalize the page data, store it, and return it.

        Args:
            page_data (object):
                The data to process.

        Returns:
            object:
            The resulting data.
        """
        self.page_data = self.normalize_page_data(page_data)

        return self.page_data



"""Add Profile.is_private.

Version Added:
    1.6
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Profile', 'is_private', models.BooleanField, initial=False)
]



"""Unit tests for reviewboard.certs.cert.CertificateFingerprints.

Version Added:
    6.0
"""

from __future__ import annotations

from cryptography import x509

from reviewboard.certs.cert import CertificateFingerprints
from reviewboard.certs.tests.testcases import (CertificateTestCase,
                                               TEST_TRUST_CERT_PEM,
                                               TEST_SHA1,
                                               TEST_SHA1_2,
                                               TEST_SHA256,
                                               TEST_SHA256_2)


class CertificateFingerprintTests(CertificateTestCase):
    """Unit tests for CertificateFingerprints.

    Version Added:
        6.0
    """

    def test_from_json(self) -> None:
        """Testing CertificateFingerprints.from_json"""
        fingerprints = CertificateFingerprints.from_json({
            'sha1': TEST_SHA1,
            'sha256': TEST_SHA256,
        })

        self.assertEqual(fingerprints.sha1, TEST_SHA1)
        self.assertEqual(fingerprints.sha256, TEST_SHA256)

    def test_from_json_with_empty(self) -> None:
        """Testing CertificateFingerprints.from_json with empty data"""
        fingerprints = CertificateFingerprints.from_json({})

        self.assertIsNone(fingerprints.sha1)
        self.assertIsNone(fingerprints.sha256)

    def test_from_json_with_bad_data(self) -> None:
        """Testing CertificateFingerprints.from_json with bad data"""
        fingerprints = CertificateFingerprints.from_json({
            'sha1': 1,
            'sha256': 2,
        })

        self.assertIsNone(fingerprints.sha1)
        self.assertIsNone(fingerprints.sha256)

    def test_from_string_with_sha1_fingerprint(self) -> None:
        """Testing CertificateFingerprints.from_string with SHA-1 fingerprint
        """
        fingerprints = CertificateFingerprints.from_string(TEST_SHA1)

        assert fingerprints is not None
        self.assertEqual(fingerprints.sha1, TEST_SHA1)
        self.assertIsNone(fingerprints.sha256)

    def test_from_string_with_sha1_fingerprint_invalid(self) -> None:
        """Testing CertificateFingerprints.from_string with invalid SHA-1
        fingerprint
        """
        self.assertIsNone(CertificateFingerprints.from_string(
            f'{TEST_SHA1[:-1]}Z'
        ))

    def test_from_string_with_sha1_string(self) -> None:
        """Testing CertificateFingerprints.from_string with SHA-1 string"""
        fingerprints = CertificateFingerprints.from_string(
            TEST_SHA1
            .replace(':', '')
            .lower()
        )

        assert fingerprints is not None
        self.assertEqual(fingerprints.sha1, TEST_SHA1)
        self.assertIsNone(fingerprints.sha256)

    def test_from_string_with_sha1_string_invalid(self) -> None:
        """Testing CertificateFingerprints.from_string with invalid SHA-1
        string
        """
        sha = TEST_SHA1.replace(':', '').lower()

        self.assertIsNone(CertificateFingerprints.from_string(
            f'{sha[:-1]}z'
        ))

    def test_from_string_with_sha256_fingerprint(self) -> None:
        """Testing CertificateFingerprints.from_string with SHA-256
        fingerprint
        """
        fingerprints = CertificateFingerprints.from_string(TEST_SHA256)

        assert fingerprints is not None
        self.assertIsNone(fingerprints.sha1)
        self.assertEqual(fingerprints.sha256, TEST_SHA256)

    def test_from_string_with_sha256_fingerprint_invalid(self) -> None:
        """Testing CertificateFingerprints.from_string with invalid SHA-256
        fingerprint
        """
        self.assertIsNone(CertificateFingerprints.from_string(
            f'{TEST_SHA256[:-1]}Z'
        ))

    def test_from_string_with_sha256_string(self) -> None:
        """Testing CertificateFingerprints.from_string with SHA-256 string"""
        fingerprints = CertificateFingerprints.from_string(
            TEST_SHA256
            .replace(':', '')
            .lower()
        )

        assert fingerprints is not None
        self.assertIsNone(fingerprints.sha1)
        self.assertEqual(fingerprints.sha256, TEST_SHA256)

    def test_from_string_with_sha256_string_invalid(self) -> None:
        """Testing CertificateFingerprints.from_string with invalid SHA-256
        string
        """
        sha = TEST_SHA256.replace(':', '').lower()

        self.assertIsNone(CertificateFingerprints.from_string(
            f'{sha[:-1]}z'
        ))

    def test_from_string_with_bad_string(self) -> None:
        """Testing CertificateFingerprints.from_string with bad string"""
        fingerprints = CertificateFingerprints.from_string('AA:BB:CC')

        self.assertIsNone(fingerprints)

    def test_from_x509_cert(self) -> None:
        """Testing CertificateFingerprints.from_x509_cert"""
        fingerprints = CertificateFingerprints.from_x509_cert(
            x509.load_pem_x509_certificate(TEST_TRUST_CERT_PEM))

        self.assertEqual(fingerprints.sha1, TEST_SHA1)
        self.assertEqual(fingerprints.sha256, TEST_SHA256)

    def test_to_json(self) -> None:
        """Testing CertificateFingerprints.to_json"""
        fingerprints = CertificateFingerprints.from_json({
            'sha1': TEST_SHA1,
            'sha256': TEST_SHA256,
        })

        self.assertEqual(fingerprints.to_json(), {
            'sha1': TEST_SHA1,
            'sha256': TEST_SHA256,
        })

    def test_to_json_with_empty(self) -> None:
        """Testing CertificateFingerprints.to_json with empty fingerprints"""
        fingerprints = CertificateFingerprints()

        self.assertEqual(fingerprints.to_json(), {})

    def test_is_empty_with_empty(self) -> None:
        """Testing CertificateFingerprints.is_empty with empty fingerprints"""
        fingerprints = CertificateFingerprints()

        self.assertTrue(fingerprints.is_empty())

    def test_is_empty_with_not_empty(self) -> None:
        """Testing CertificateFingerprints.is_empty with fingerprints"""
        fingerprints = CertificateFingerprints(sha1=TEST_SHA1)
        self.assertFalse(fingerprints.is_empty())

        fingerprints = CertificateFingerprints(sha256=TEST_SHA256)
        self.assertFalse(fingerprints.is_empty())

    def test_matches_with_empty(self) -> None:
        """Testing CertificateFingerprints.matches with empty"""
        fingerprints1 = CertificateFingerprints()
        fingerprints2 = CertificateFingerprints()

        self.assertFalse(fingerprints1.matches(fingerprints2))
        self.assertFalse(fingerprints2.matches(fingerprints1))

    def test_matches_with_sha1_only_match(self) -> None:
        """Testing CertificateFingerprints.matches with SHA1 only and match"""
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1)
        fingerprints2 = CertificateFingerprints(sha1=TEST_SHA1)

        self.assertTrue(fingerprints1.matches(fingerprints2))
        self.assertTrue(fingerprints2.matches(fingerprints1))

    def test_matches_with_sha1_only_no_match(self) -> None:
        """Testing CertificateFingerprints.matches with SHA1 only and no match
        """
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1)
        fingerprints2 = CertificateFingerprints(sha1=TEST_SHA1_2)

        self.assertFalse(fingerprints1.matches(fingerprints2))
        self.assertFalse(fingerprints2.matches(fingerprints1))

    def test_matches_with_sha256_only_match(self) -> None:
        """Testing CertificateFingerprints.matches with SHA256 only and match
        """
        fingerprints1 = CertificateFingerprints(sha256=TEST_SHA256)
        fingerprints2 = CertificateFingerprints(sha256=TEST_SHA256)

        self.assertTrue(fingerprints1.matches(fingerprints2))
        self.assertTrue(fingerprints2.matches(fingerprints1))

    def test_matches_with_sha256_only_no_match(self) -> None:
        """Testing CertificateFingerprints.matches with SHA256 only and no
        match
        """
        fingerprints1 = CertificateFingerprints(sha256=TEST_SHA256)
        fingerprints2 = CertificateFingerprints(sha256=TEST_SHA256_2)

        self.assertFalse(fingerprints1.matches(fingerprints2))
        self.assertFalse(fingerprints2.matches(fingerprints1))

    def test_matches_with_sha_type_mismatch(self) -> None:
        """Testing CertificateFingerprints.matches with mismatch in available
        fingerprint types
        """
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1)
        fingerprints2 = CertificateFingerprints(sha256=TEST_SHA256)

        self.assertFalse(fingerprints1.matches(fingerprints2))
        self.assertFalse(fingerprints2.matches(fingerprints1))

    def test_matches_with_match(self) -> None:
        """Testing CertificateFingerprints.matches with match"""
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)
        fingerprints2 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)

        self.assertTrue(fingerprints1.matches(fingerprints2))
        self.assertTrue(fingerprints2.matches(fingerprints1))

    def test_matches_with_no_match(self) -> None:
        """Testing CertificateFingerprints.matches with no match"""
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)
        fingerprints2 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256_2)

        self.assertFalse(fingerprints1.matches(fingerprints2))
        self.assertFalse(fingerprints2.matches(fingerprints1))

    def test_eq_with_equal(self) -> None:
        """"Testing CertificateFingerprints.__eq__ with objects equal"""
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)
        fingerprints2 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)

        self.assertEqual(fingerprints1, fingerprints2)

    def test_eq_with_not_equal(self) -> None:
        """"Testing CertificateFingerprints.__eq__ with objects not equal"""
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)
        fingerprints2 = CertificateFingerprints(sha1=TEST_SHA1)

        self.assertNotEqual(fingerprints1, fingerprints2)

    def test_eq_with_wrong_object(self) -> None:
        """"Testing CertificateFingerprints.__eq__ with wrong object type"""
        fingerprints1 = CertificateFingerprints(sha1=TEST_SHA1,
                                                sha256=TEST_SHA256)

        self.assertNotEqual(fingerprints1, 123)
        self.assertNotEqual(fingerprints1, None)



"""Unit tests for reviewboard.site.middleware.LocalSiteMiddleware."""

from django.http import HttpRequest, HttpResponse

from reviewboard.site.middleware import LocalSiteMiddleware
from reviewboard.site.models import LocalSite
from reviewboard.testing.testcase import TestCase


class LocalSiteMiddlewareTests(TestCase):
    """Unit tests for reviewboard.site.middleware.LocalSiteMiddleware."""

    def setUp(self):
        super().setUp()

        self.middleware = LocalSiteMiddleware(lambda: HttpResponse(''))

    def test_request_local_site_empty(self):
        """Testing LocalSiteMiddleware's request.local_site with no LocalSite
        """
        request = HttpRequest()
        self.middleware.process_view(request=request, view_func=None,
                                     view_args=None, view_kwargs={})

        self.assertTrue(hasattr(request, '_local_site_name'))
        self.assertTrue(hasattr(request, 'local_site'))
        self.assertIsNone(request._local_site_name)
        self.assertIsNone(request.local_site)

    def test_request_local_site_not_empty(self):
        """Testing LocalSiteMiddleware's request.local_site with a LocalSite"""
        local_site = LocalSite.objects.create(name='test-site')

        request = HttpRequest()
        self.middleware.process_view(
            request=request,
            view_func=None,
            view_args=None,
            view_kwargs={
                'local_site_name': local_site.name,
            })

        self.assertTrue(hasattr(request, '_local_site_name'))
        self.assertTrue(hasattr(request, 'local_site'))
        self.assertEqual(request._local_site_name, 'test-site')
        self.assertEqual(request.local_site, local_site)



"""Unit tests for NISBackend."""

from unittest import SkipTest

import kgb
from django.contrib.auth.models import User

from reviewboard.accounts.backends import NISBackend
from reviewboard.testing import TestCase


class NISBackendTests(kgb.SpyAgency, TestCase):
    """Unit tests for NISBackend."""

    NIS_SITECONFIG_SETTINGS = {
        'auth_nis_email_domain': 'example.com',
    }

    def setUp(self):
        super(NISBackendTests, self).setUp()

        self.backend = NISBackend()

        if self.backend.nis is None:
            raise SkipTest('nis is not available on this build of Python.')

    def test_authenticate_with_valid_user(self):
        """Testing NISBackend.authenticate with valid user credentials"""
        self.spy_on(self.backend._nis_get_passwd, lambda _self, username: (
            'test-user', '6a6Sl/u3EOBfo', 100, 100, 'Test User,Foo Bar',
            '/home/test-user', '/bin/sh'))

        with self.siteconfig_settings(self.NIS_SITECONFIG_SETTINGS):
            user = self.backend.authenticate(request=None,
                                             username='test-user',
                                             password='test-pass')

        # Use a plain assert to help the type checker.
        assert user is not None

        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)
        self.assertSpyCalledWith(self.backend._nis_get_passwd, 'test-user')

    def test_authenticate_with_invalid_credentials(self):
        """Testing NISBackend.authenticate with bad user credentials"""

        self.spy_on(self.backend._nis_get_passwd, lambda _self, username: (
            'test-user', 'foo', 100, 100, 'Test User',
            '/home/test-user', '/bin/sh'))

        user = self.backend.authenticate(request=None,
                                         username='test-user',
                                         password='test-pass')
        self.assertIsNone(user)

    def test_authenticate_with_invalid_user(self):
        """Testing NISBackend.authenticate with bad user credentials"""
        backend = self.backend

        self.spy_on(backend._nis_get_passwd,
                    op=kgb.SpyOpRaise(backend.nis.error()))

        user = backend.authenticate(request=None,
                                    username='test-user',
                                    password='test-pass')
        self.assertIsNone(user)

    def test_get_or_create_user_with_user_in_db(self):
        """Testing NISBackend.get_or_create_user with user in database"""
        user = User.objects.create(username='test-user')

        self.assertEqual(
            self.backend.get_or_create_user(username='test-user'),
            user)

    def test_get_or_create_user_with_user_in_nis(self):
        """Testing NISBackend.get_or_create_user with user in NIS"""
        self.spy_on(self.backend._nis_get_passwd, lambda _self, username: (
            'test-user', '6a6Sl/u3EOBfo', 100, 100, 'Test User',
            '/home/test-user', '/bin/sh'))

        user = User.objects.create(username='test-user')

        self.assertEqual(
            self.backend.get_or_create_user(username='test-user'),
            user)

    def test_get_or_create_user_with_user_and_passwd(self):
        """Testing NISBackend.get_or_create_user with explicit passwd value"""
        self.spy_on(self.backend._nis_get_passwd, lambda _self, username: None)

        with self.siteconfig_settings(self.NIS_SITECONFIG_SETTINGS):
            user = self.backend.get_or_create_user(
                username='test-user',
                passwd=('test-user', '6a6Sl/u3EOBfo', 100, 100,
                        'Test User,Foo Bar', '/home/test-user', '/bin/sh'))

        # Use a plain assert to help the type checker.
        assert user is not None

        self.assertEqual(user.username, 'test-user')
        self.assertEqual(user.first_name, 'Test')
        self.assertEqual(user.last_name, 'User')
        self.assertEqual(user.email, 'test-user@example.com')
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)
        self.assertSpyNotCalled(self.backend._nis_get_passwd)



"""Unit tests for ReviewRequestPageData."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone as tz
from typing import Callable, Optional, Sequence, TYPE_CHECKING

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.db.models import Q
from django.template import RequestContext
from django.test.client import RequestFactory
from django.utils import timezone

from reviewboard.accounts.models import Profile
from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.diffviewer.models import DiffSet, FileDiff
from reviewboard.hostingsvcs.github import GitHub
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.reviews.detail import (BaseReviewRequestPageEntry,
                                        ChangeEntry,
                                        InitialStatusUpdatesEntry,
                                        ReviewEntry,
                                        ReviewRequestEntry,
                                        ReviewRequestPageData)
from reviewboard.reviews.models import (BaseComment,
                                        Review,
                                        ReviewRequest,
                                        ReviewRequestDraft,
                                        Screenshot,
                                        StatusUpdate)
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQuery

    from reviewboard.reviews.models import (Comment,
                                            FileAttachmentComment,
                                            GeneralComment,
                                            ScreenshotComment)


class ReviewRequestPageDataTests(TestCase):
    """Unit tests for ReviewRequestPageData."""

    fixtures = ['test_scmtools', 'test_users']

    ######################
    # Instance variables #
    ######################

    #: The test comments created.
    #:
    #: This covers all comment types.
    #:
    #: Version Added:
    #:     8.0
    all_comments: list[BaseComment]

    #: The test change descriptions created.
    #:
    #: Version Added:
    #:     8.0
    changedescs: list[ChangeDescription]

    #: The test diff comments created.
    #:
    #: Version Added:
    #:     8.0
    diff_comments: list[Comment]

    #: The test DiffSets created.
    #:
    #: Version Added:
    #:     8.0
    diffsets: list[DiffSet]

    #: The test draft created.
    draft: Optional[ReviewRequestDraft]

    #: The test file attachment comments created.
    #:
    #: Version Added:
    #:     8.0
    file_attachment_comments: list[FileAttachmentComment]

    #: The test file attachments created.
    #:
    #: Version Added:
    #:     8.0
    file_attachments: list[FileAttachment]

    #: The test FileDiffs created.
    #:
    #: Version Added:
    #:     8.0
    filediffs: list[FileDiff]

    #: The test general comments created.
    #:
    #: Version Added:
    #:     8.0
    general_comments: list[GeneralComment]

    #: The test inactive file attachments created.
    #:
    #: Version Added:
    #:     8.0
    inactive_screenshots: list[Screenshot]

    #: The test inactive screenshots created.
    #:
    #: Version Added:
    #:     8.0
    inactive_file_attachments: list[FileAttachment]

    #: The test issue counts created.
    #:
    #: Version Added:
    #:     8.0
    issue_counts: dict[str, int]

    #: The test reviews created.
    #:
    #: Version Added:
    #:     8.0
    reviews: list[Review]

    #: The test screenshot comments created.
    #:
    #: Version Added:
    #:     8.0
    screenshot_comments: list[ScreenshotComment]

    #: The test screenshots created.
    #:
    #: Version Added:
    #:     8.0
    screenshots: list[Screenshot]

    #: The test status updates created.
    #:
    #: Version Added:
    #:     8.0
    status_updates: list[StatusUpdate]

    def test_query_data_pre_etag(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag"""
        self._populate_review_request()

        self._test_query_data_pre_etag_with(
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_changedescs': 'INNER JOIN',
                    },
                    'model': ChangeDescription,
                    'num_joins': 1,
                    'tables': {
                        'changedescs_changedescription',
                        'reviews_reviewrequest_changedescs',
                    },
                    'where': (
                        Q(review_request__id=1) &
                        Q(public=True)
                    ),
                },
                {
                    'model': ReviewRequestDraft,
                    'where': Q(review_request=self.review_request),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(DiffSet.objects.all())),
                },
                {
                    'model': StatusUpdate,
                    'order_by': ('summary',),
                    'where': Q(review_request=self.review_request),
                },
            ],
            expect_diffs=True,
            expect_reviews=True,
            expect_changedescs=True,
            expect_draft=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_only_review_request_entry(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        ReviewRequestEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_pre_etag_with(
            entry_classes=[ReviewRequestEntry],
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'model': ReviewRequestDraft,
                    'where': Q(review_request=self.review_request),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(DiffSet.objects.all())),
                },
            ],
            expect_diffs=True,
            expect_reviews=True,
            expect_draft=True)

    def test_query_data_pre_etag_with_only_initial_status_updates_entry(
        self,
    ) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        InitialStatusUpdatesEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_pre_etag_with(
            entry_classes=[InitialStatusUpdatesEntry],
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(DiffSet.objects.all())),
                },
                {
                    'model': StatusUpdate,
                    'order_by': ('summary',),
                    'where': Q(review_request=self.review_request),
                },
            ],
            expect_diffs=True,
            expect_reviews=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_only_review_entry(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        ReviewEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_pre_etag_with(
            entry_classes=[ReviewEntry],
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=self.diffsets),
                },
            ],
            expect_diffs=True,
            expect_reviews=True)

    def test_query_data_pre_etag_with_only_change_entry(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        ChangeEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_pre_etag_with(
            entry_classes=[ChangeEntry],
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_changedescs': 'INNER JOIN',
                    },
                    'model': ChangeDescription,
                    'num_joins': 1,
                    'tables': {
                        'changedescs_changedescription',
                        'reviews_reviewrequest_changedescs',
                    },
                    'where': (
                        Q(review_request__id=self.review_request.pk) &
                        Q(public=True)
                    ),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(DiffSet.objects.all())),
                },
                {
                    'model': StatusUpdate,
                    'order_by': ('summary',),
                    'where': Q(review_request=self.review_request),
                },
            ],
            expect_diffs=True,
            expect_reviews=True,
            expect_changedescs=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_no_changedescs(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with no
        change descriptions
        """
        self._populate_review_request(enable_changedescs=False)

        self._test_query_data_pre_etag_with(
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_changedescs': 'INNER JOIN',
                    },
                    'model': ChangeDescription,
                    'num_joins': 1,
                    'tables': {
                        'changedescs_changedescription',
                        'reviews_reviewrequest_changedescs',
                    },
                    'where': (
                        Q(review_request__id=self.review_request.pk) &
                        Q(public=True)
                    ),
                },
                {
                    'model': ReviewRequestDraft,
                    'where': Q(review_request=self.review_request),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(DiffSet.objects.all())),
                },
                {
                    'model': StatusUpdate,
                    'order_by': ('summary',),
                    'where': Q(review_request=self.review_request),
                },
            ],
            expect_diffs=True,
            expect_draft=True,
            expect_reviews=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_no_diffs(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with no diffs"""
        self._populate_review_request(enable_diffs=False)

        self._test_query_data_pre_etag_with(
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_changedescs': 'INNER JOIN',
                    },
                    'model': ChangeDescription,
                    'num_joins': 1,
                    'tables': {
                        'changedescs_changedescription',
                        'reviews_reviewrequest_changedescs',
                    },
                    'where': (
                        Q(review_request__id=self.review_request.pk) &
                        Q(public=True)
                    ),
                },
                {
                    'model': ReviewRequestDraft,
                    'where': Q(review_request=self.review_request),
                },
                {
                    'model': StatusUpdate,
                    'order_by': ('summary',),
                    'where': Q(review_request=self.review_request),
                },
            ],
            expect_changedescs=True,
            expect_draft=True,
            expect_reviews=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_no_reviews(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag with no reviews"""
        self._populate_review_request(enable_reviews=False)

        self._test_query_data_pre_etag_with(
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_changedescs': 'INNER JOIN',
                    },
                    'model': ChangeDescription,
                    'num_joins': 1,
                    'tables': {
                        'changedescs_changedescription',
                        'reviews_reviewrequest_changedescs',
                    },
                    'where': (
                        Q(review_request__id=self.review_request.pk) &
                        Q(public=True)
                    ),
                },
                {
                    'model': ReviewRequestDraft,
                    'where': Q(review_request=self.review_request),
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(DiffSet.objects.all())),
                },
                {
                    'model': StatusUpdate,
                    'order_by': ('summary',),
                    'where': Q(review_request=self.review_request),
                },
                {
                    'model': User,
                    'where': Q(pk__in=(3,)),
                },
            ],
            expect_diffs=True,
            expect_draft=True,
            expect_changedescs=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_draft_diffsets(self) -> None:
        """Testing ReviewRequestPageData.query_data_pre_etag includes
        draft diffsets
        """
        self._populate_review_request()

        # Create a draft diffset.
        review_request = self.review_request
        draft_diffset = self.create_diffset(review_request, draft=True)

        self.diffsets.append(draft_diffset)

        self._test_query_data_pre_etag_with(
            entry_classes=[ReviewRequestEntry],
            expected_queries=lambda: [
                {
                    'model': Review,
                    'order_by': ('-timestamp',),
                    'select_related': {'user'},
                    'where': (
                        Q(review_request=self.review_request) &
                        (Q(public=True) |
                         Q(user_id=2))
                    ),
                },
                {
                    'model': ReviewRequest,
                    'where': Q(id=review_request.pk),
                },
                {
                    'model': User,
                    'where': Q(id=review_request.owner.pk)
                },
                {
                    'model': DiffSet,
                    'where': Q(history__pk=1),
                },
                {
                    'model': FileDiff,
                    'where': Q(diffset__in=list(
                        DiffSet.objects.filter(history__pk=1))),
                },
            ],
            expect_diffs=True,
            expect_reviews=True,
            expect_draft=True)

    def test_query_data_post_etag(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag"""
        self._populate_review_request()

        self._test_query_data_post_etag_with(
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_file_attachments',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_inactive_file_attachments',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_draft=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_review_request_entry(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with only
        ReviewRequestEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_post_etag_with(
            entry_classes=[ReviewRequestEntry],
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_file_attachments',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_inactive_file_attachments',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_draft=True,
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_initial_status_updates_entry(
        self,
    ) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with only
        InitialStatusUpdatesEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_post_etag_with(
            entry_classes=[InitialStatusUpdatesEntry],
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequest_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequest_file_attachments',
                    },
                    'where': Q(review_request__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequest_inactive_file_attachments',
                    },
                    'where': Q(inactive_review_request__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(review_request__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_review_request__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_review_entry(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with only
        ReviewEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_post_etag_with(
            entry_classes=[ReviewEntry],
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequest_file_attachments',
                    },
                    'where': Q(review_request__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequest_inactive_file_attachments',
                    },
                    'where': Q(inactive_review_request__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(review_request__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_review_request__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_change_entry(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with only
        ChangeEntry in entry_classes
        """
        self._populate_review_request()

        self._test_query_data_post_etag_with(
            entry_classes=[ChangeEntry],
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequest_file_attachments',
                    },
                    'where': Q(review_request__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequest_inactive_file_attachments',
                    },
                    'where': Q(inactive_review_request__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(review_request__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequest_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequest_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_review_request__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_no_comments(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with no
        comments
        """
        self._populate_review_request(
            enable_diff_comments=False,
            enable_file_attachment_comments=False,
            enable_general_comments=False,
            enable_screenshot_comments=False)

        self._test_query_data_post_etag_with(
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_file_attachments',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_inactive_file_attachments',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_draft=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_no_file_attachments(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with no
        file attachments
        """
        self._populate_review_request(enable_file_attachments=False)

        self._test_query_data_post_etag_with(
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_screenshots': 'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_screenshots':
                            'INNER JOIN',
                    },
                    'model': Screenshot,
                    'num_joins': 1,
                    'tables': {
                        'reviews_reviewrequestdraft_inactive_screenshots',
                        'reviews_screenshot',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.screenshot_comments.through,
                    'order_by': ('screenshotcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_screenshots=True,
            expect_draft=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_draft_diff_comments(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with draft
        diff comments
        """
        self._populate_review_request(enable_file_attachments=False,
                                      enable_screenshots=False)

        # Create a draft diffset.
        review_request = self.review_request
        draft_diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(draft_diffset)

        self.diffsets.append(draft_diffset)

        # Create a draft comment owned by the requesting user.
        review = self.create_review(review_request, user='doc')
        comment = self.create_diff_comment(review, filediff)

        self.reviews.append(review)
        self.all_comments.append(comment)

        # Create a draft comment owned by a different user.
        #
        # The review won't appear in the page data, but the comment will. The
        # comment won't actually appear on the page, it gets filtered out in
        # ReviewsDiffViewerView._get_comments_hint().
        review_admin = self.create_review(review_request, user='admin')
        other_comment = self.create_diff_comment(review, filediff)

        self.all_comments.append(other_comment)

        review_pks = [10, 9, 8, 7, 6, 5, 4, 3, 2, 11, 1]

        self._test_query_data_post_etag_with(
            expected_queries=lambda: [
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 11, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=review_pks),
                },
            ],
            expect_reviews=True,
            expect_draft=True,
            expect_comments=True,
            expect_issues=True)
        self.assertNotIn(review_admin.pk, review_pks)

    def test_query_data_post_etag_with_no_screenshots(self) -> None:
        """Testing ReviewRequestPageData.query_data_post_etag with no
        screenshots
        """
        self._populate_review_request(enable_screenshots=False)

        self._test_query_data_post_etag_with(
            expected_queries=lambda: [
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_file_attachments',
                    },
                    'where': Q(drafts__id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=1),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=2),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=3),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=4),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=5),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=6),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=7),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=8),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=9),
                },
                {
                    'model': FileAttachmentHistory,
                    'where': Q(id=10),
                },
                {
                    'join_types': {
                        'reviews_reviewrequestdraft_inactive_file_attachments':
                            'INNER JOIN',
                    },
                    'model': FileAttachment,
                    'num_joins': 1,
                    'tables': {
                        'attachments_fileattachment',
                        'reviews_reviewrequestdraft_inactive_file_attachments',
                    },
                    'where': Q(inactive_drafts__id=1),
                },
                {
                    'model': Review.general_comments.through,
                    'order_by': ('generalcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.file_attachment_comments.through,
                    'order_by': ('fileattachmentcomment__timestamp',),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
                {
                    'model': Review.comments.through,
                    'order_by': ('comment__filediff',
                                 'comment__first_line',
                                 'comment__timestamp'),
                    'select_related': True,
                    'where': Q(review__in=[10, 9, 8, 7, 6, 5, 4, 3, 2, 1]),
                },
            ],
            expect_reviews=True,
            expect_file_attachments=True,
            expect_draft=True,
            expect_comments=True,
            expect_issues=True)

    def test_get_entries(self) -> None:
        """Testing ReviewRequestPageData.get_entries"""
        self._populate_review_request()

        data = self._build_data()
        data.query_data_pre_etag()
        data.query_data_post_etag()

        queries: list[ExpectedQuery] = [
            {
                'join_types': {
                    'reviews_reviewrequest_changedescs': 'INNER JOIN',
                },
                'model': ChangeDescription,
                'num_joins': 1,
                'order_by': ('-pk',),
                'tables': {
                    'changedescs_changedescription',
                    'reviews_reviewrequest_changedescs',
                },
                'where': (
                    Q(review_request__id=1) &
                    Q(pk__lt=3)
                ),
            },
            {
                'model': ChangeDescription,
                'type': 'UPDATE',
                'where': Q(pk=3),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_changedescs': 'INNER JOIN',
                },
                'model': ChangeDescription,
                'num_joins': 1,
                'order_by': ('-pk',),
                'tables': {
                    'changedescs_changedescription',
                    'reviews_reviewrequest_changedescs',
                },
                'where': (
                    Q(review_request__id=1) &
                    Q(pk__lt=2)
                ),
            },
            {
                'model': ChangeDescription,
                'type': 'UPDATE',
                'where': Q(pk=2),
            },
        ]

        with assert_queries(queries):
            entries = data.get_entries()

        initial_entries = entries['initial']
        main_entries = entries['main']
        reviews = self.reviews
        changedescs = self.changedescs

        self.assertEqual(len(initial_entries), 1)
        self.assertEqual(len(main_entries), len(reviews) + len(changedescs))

        entry = initial_entries[0]
        self.assertIsInstance(entry, InitialStatusUpdatesEntry)

        for i in range(len(reviews)):
            entry = main_entries[i]
            assert isinstance(entry, ReviewEntry)
            self.assertEqual(entry.review, reviews[i])

        start = len(reviews)

        for i in range(len(changedescs)):
            entry = main_entries[start + i]
            assert isinstance(entry, ChangeEntry)
            self.assertEqual(entry.changedesc, changedescs[i])

    def test_rendering(self) -> None:
        """Testing ReviewRequestPageData entries rendering"""
        # We're testing a render across all the entries in order to see
        # if we're generating too many SQL queries.
        self._populate_review_request()

        data = self._build_data()
        data.query_data_pre_etag()
        data.query_data_post_etag()

        request = data.request
        context = RequestContext(request)

        owner = self.review_request.owner
        reviewer = self.reviews[0].user
        viewer = request.user

        # Ensure users have profiles for the test, to avoid excess queries.
        owner.get_profile()
        reviewer.get_profile()
        viewer.get_profile()

        entries = data.get_entries()

        queries: list[ExpectedQuery] = [
            {
                'model': Profile,
                'where': Q(user=reviewer)
            },
        ]

        with assert_queries(queries):
            for key in ('initial', 'main'):
                for entry in entries[key]:
                    entry.render_to_string(request=request,
                                           context=context)

    def _build_data(
        self,
        entry_classes: Optional[
            Sequence[type[BaseReviewRequestPageEntry]]
        ] = None,
    ) -> ReviewRequestPageData:
        """Build the data to test against.

        This will build the page data to test against. It must be run
        after populating review request data.

        Args:
            entry_classes (list of type, optional):
                The list of entry classes available for the page data.

        Returns:
            reviewboard.reviews.detail.ReviewRequestPageData:
            The resulting page data.
        """
        request = RequestFactory().get('/r/1/')
        request.user = self.review_request.submitter

        return ReviewRequestPageData(review_request=self.review_request,
                                     request=request,
                                     entry_classes=entry_classes)

    def _test_query_data_pre_etag_with(
        self,
        *,
        expected_queries: Callable[[], Sequence[ExpectedQuery]],
        entry_classes: Optional[
            Sequence[type[BaseReviewRequestPageEntry]]
        ] = None,
        expect_changedescs: bool = False,
        expect_diffs: bool = False,
        expect_draft: bool = False,
        expect_reviews: bool = False,
        expect_status_updates: bool = False,
    ) -> None:
        """Perform a test of state from pre-ETag calculations.

        Version Changed:
            8.0:
            Added the following argument: ``expected_queries``,
            ``expect_diffs``.

        Args:
            expected_queries (callable):
                A function returning a list of expected queries.

                Version Added:
                    8.0

            entry_classes (list of type, optional):
                The list of entry classes available for the page data.

            expect_changedescs (bool, optional):
                Whether to expect change descriptions in the results.

            expect_diffs (bool, optional):
                Whether to expect diffs in the results.

                Version Added:
                    8.0

            expect_draft (bool, optional):
                Whether to expect a review request draft in the results.

            expect_reviews (bool, optional):
                Whether to expect reviews in the results.

            expect_status_updates (bool, optional):
                Whether to expect status updates in the results.
        """
        data = self._build_data(entry_classes=entry_classes)

        with assert_queries(expected_queries()):
            data.query_data_pre_etag()

        if expect_reviews:
            self.assertEqual(data.reviews, self.reviews[::-1])
            self.assertEqual(data.latest_review_timestamp,
                             self.reviews[-1].timestamp)
        else:
            self.assertEqual(data.reviews, [])
            self.assertEqual(data.latest_review_timestamp,
                             datetime.fromtimestamp(0, tz.utc))

        if expect_diffs:
            self.assertEqual(data.diffsets, self.diffsets)
            self.assertEqual(
                data.diffsets_by_id,
                {
                    pk: diffset
                    for pk, diffset in enumerate(self.diffsets, start=1)
                })
        else:
            self.assertEqual(data.diffsets, [])
            self.assertEqual(data.diffsets_by_id, {})

        if expect_changedescs:
            self.assertEqual(data.changedescs, self.changedescs[::-1])
            self.assertEqual(data.latest_changedesc_timestamp,
                             self.changedescs[-1].timestamp)
        else:
            self.assertEqual(data.changedescs, [])
            self.assertIsNone(data.latest_changedesc_timestamp)

        if expect_draft:
            self.assertEqual(data.draft, self.draft)
        else:
            self.assertIsNone(data.draft)

        if expect_status_updates:
            self.assertEqual(data.all_status_updates, self.status_updates)
        else:
            self.assertEqual(data.all_status_updates, [])

    def _test_query_data_post_etag_with(
        self,
        *,
        expected_queries: Callable[[], Sequence[ExpectedQuery]],
        entry_classes: Optional[
            Sequence[type[BaseReviewRequestPageEntry]]
        ] = None,
        expect_comments: bool = False,
        expect_draft: bool = False,
        expect_file_attachments: bool = False,
        expect_issues: bool = False,
        expect_reviews: bool = False,
        expect_screenshots: bool = False,
    ) -> None:
        """Perform a test of state from post-ETag calculations.

        Version Changed:
            8.0:
            Added the following argument: ``expected_queries``,

        Args:
            expected_queries (callable):
                A function returning a list of expected queries.

                Version Added:
                    8.0

            entry_classes (list of type, optional):
                The list of entry classes available for the page data.

            expect_comments (bool, optional):
                Whether to expect comments in the results.

            expect_draft (bool, optional):
                Whether to expect a review request draft in the results.

            expect_file_attachments (bool, optional):
                Whether to expect file attachments in the results.

            expect_issues (bool, optional):
                Whether to expect filed issues in the results.

            expect_reviews (bool, optional):
                Whether to expect reviews in the results.

            expect_screenshots (bool, optional):
                Whether to expect screenshots in the results.
        """
        data = self._build_data(entry_classes=entry_classes)

        data.query_data_pre_etag()

        with assert_queries(expected_queries()):
            data.query_data_post_etag()

        if expect_reviews:
            self.assertEqual(
                data.reviews_by_id,
                {
                    pk: review
                    for pk, review in enumerate(self.reviews, start=1)
                })
        else:
            self.assertEqual(data.reviews_by_id, {})

        if expect_draft:
            self.assertEqual(data.review_request_details, self.draft)
        else:
            self.assertEqual(data.review_request_details,
                             self.review_request)

        if expect_file_attachments:
            all_expected_file_attachments = \
                self.file_attachments + self.inactive_file_attachments

            self.assertEqual(data.active_file_attachments,
                             self.file_attachments)
            self.assertEqual(data.all_file_attachments,
                             all_expected_file_attachments)
            self.assertEqual(
                data.file_attachments_by_id,
                {
                    pk: file_attachment
                    for pk, file_attachment in enumerate(
                        all_expected_file_attachments,
                        start=1)
                })
        else:
            self.assertEqual(data.active_file_attachments, [])
            self.assertEqual(data.all_file_attachments, [])
            self.assertEqual(data.file_attachments_by_id, {})

        if expect_screenshots:
            all_expected_screenshots = \
                self.screenshots + self.inactive_screenshots

            self.assertEqual(data.active_screenshots, self.screenshots)
            self.assertEqual(data.all_screenshots, all_expected_screenshots)
            self.assertEqual(
                data.screenshots_by_id,
                {
                    pk: screenshot
                    for pk, screenshot in enumerate(
                        all_expected_screenshots,
                        start=1)
                })
        else:
            self.assertEqual(data.active_screenshots, [])
            self.assertEqual(data.all_screenshots, [])
            self.assertEqual(data.screenshots_by_id, {})

        if expect_comments:
            self.assertEqual(data.all_comments, self.all_comments)
        else:
            self.assertEqual(data.all_comments, [])

        if expect_issues:
            self.assertEqual(
                data.issues,
                [
                    comment
                    for comment in self.all_comments
                    if comment.issue_opened
                ])
            self.assertEqual(data.issue_counts, self.issue_counts)
        else:
            self.assertEqual(data.issues, [])
            self.assertEqual(data.issue_counts, {})

    def _populate_review_request(
        self,
        *,
        enable_changedescs: bool = True,
        enable_diff_comments: bool = True,
        enable_diffs: bool = True,
        enable_draft: bool = True,
        enable_file_attachment_comments: bool = True,
        enable_file_attachments: bool = True,
        enable_general_comments: bool = True,
        enable_reviews: bool = True,
        enable_screenshot_comments: bool = True,
        enable_screenshots: bool = True,
        enable_status_updates: bool = True,
    ) -> None:
        """Populate review request state to test against.

        Version Changed:
            8.0:
            Added the following arguments: ``enable_changedescs``,
            ``enable_diff_comments``, ``enable_diffs``, ``enable_draft``,
            ``enable_file_attachment_comments``, ``enable_file_attachments``,
            ``enable_general_comments``, ``enable_reviews``,
            ``enable_screenshot_comments``, ``enable_screenshots``,
            ``enable_status_updates``.

        Args:
            enable_changedescs (bool, optional):
                Whether to enable change descriptions in the database.

                Version Added:
                    8.0

            enable_diff_comments (bool, optional):
                Whether to enable diff comments in the database.

                Version Added:
                    8.0

            enable_diffs (bool, optional):
                Whether to enable diffs in the database.

                Version Added:
                    8.0

            enable_draft (bool, optional):
                Whether to enable a draft for the revie request in the
                database.

            enable_file_attachment_comments (bool, optional):
                Whether to enable file attachment comments in the database.

                Version Added:
                    8.0

            enable_file_attachments (bool, optional):
                Whether to enable file attachments in the database.

                Version Added:
                    8.0

            enable_general_comments (bool, optional):
                Whether to enable general comments in the database.

                Version Added:
                    8.0

            enable_reviews (bool, optional):
                Whether to enable reviews in the database.

                Version Added:
                    8.0

            enable_screenshot_comments (bool, optional):
                Whether to enable screenshot comments in the database.

                Version Added:
                    8.0

            enable_screenshots (bool, optional):
                Whether to enable screenshots in the database.

                Version Added:
                    8.0

            enable_status_updates (bool, optional):
                Whether to enable Status Updates in the database.

                Version Added:
                    8.0
        """
        now = timezone.now()
        days_offset: int = 0

        # Create the review request, diffs, attachments, and screenshots.
        if enable_diffs:
            account = HostingServiceAccount.objects.create(
                service_name=GitHub.hosting_service_id,
                username='foo')
            repository = self.create_repository(hosting_account=account)
        else:
            repository = None

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.review_request = review_request

        user = review_request.owner

        #: Create the diff data.
        diffsets: list[DiffSet] = []
        filediffs: list[FileDiff] = []

        if enable_diffs:
            diffsets = [
                self.create_diffset(review_request)
                for i in range(10)
            ]

            filediffs = [
                self.create_filediff(diffset)
                for diffset in diffsets
            ]

        self.diffsets = diffsets
        self.filediffs = filediffs

        # Create the file attachments.
        file_attachments: list[FileAttachment] = []
        inactive_file_attachments: list[FileAttachment] = []

        if enable_file_attachments:
            file_attachments = [
                self.create_file_attachment(review_request)
                for i in range(10)
            ]

            inactive_file_attachments = [
                self.create_file_attachment(review_request, active=False)
                for i in range(5)
            ]

        self.file_attachments = file_attachments
        self.inactive_file_attachments = inactive_file_attachments

        # Create the screenshots.
        screenshots: list[Screenshot] = []
        inactive_screenshots: list[Screenshot] = []

        if enable_screenshots:
            screenshots = [
                self.create_screenshot(review_request)
                for i in range(10)
            ]

            inactive_screenshots = [
                self.create_screenshot(review_request, active=False)
                for i in range(5)
            ]

        self.screenshots = screenshots
        self.inactive_screenshots = inactive_screenshots

        # Create a draft for this review request.
        if enable_draft:
            draft = ReviewRequestDraft.create(review_request)
        else:
            draft = None

        self.draft = draft

        # Create some reviews.
        diff_comments: list[Comment] = []
        general_comments: list[GeneralComment] = []
        screenshot_comments: list[ScreenshotComment] = []
        file_attachment_comments: list[FileAttachmentComment] = []
        reviews: list[Review] = []
        issue_counts = {
            'dropped': 0,
            'open': 0,
            'resolved': 0,
            'total': 0,
            'verifying': 0,
        }

        if enable_reviews:
            # We'll create alternating pairs of reviews with two different
            # sets of state.
            for i in range(0, 10, 2):
                # First review in the pair.
                timestamp = now + timedelta(days=days_offset)
                review = self.create_review(
                    review_request,
                    timestamp=timestamp,
                    publish=True)
                reviews.append(review)
                days_offset += 1

                if enable_general_comments:
                    general_comments.append(
                        self.create_general_comment(
                            review,
                            issue_opened=True,
                            issue_status=BaseComment.OPEN,
                            timestamp=timestamp))
                    issue_counts['open'] += 1
                    issue_counts['total'] += 1

                if enable_diffs and enable_diff_comments:
                    diff_comments.append(
                        self.create_diff_comment(
                            review,
                            filediffs[i],
                            issue_opened=True,
                            issue_status=BaseComment.RESOLVED,
                            timestamp=timestamp))
                    issue_counts['resolved'] += 1
                    issue_counts['total'] += 1

                if enable_file_attachments and enable_file_attachment_comments:
                    file_attachment_comments.append(
                        self.create_file_attachment_comment(
                            review,
                            file_attachments[i],
                            issue_opened=True,
                            issue_status=BaseComment.DROPPED,
                            timestamp=timestamp))
                    issue_counts['dropped'] += 1
                    issue_counts['total'] += 1

                if enable_screenshots and enable_screenshot_comments:
                    screenshot_comments.append(
                        self.create_screenshot_comment(
                            review,
                            screenshots[i],
                            issue_opened=False,
                            timestamp=timestamp))

                # Second review in the pair.
                timestamp = now + timedelta(days=days_offset)
                review = self.create_review(
                    review_request,
                    timestamp=timestamp,
                    publish=True)
                reviews.append(review)
                days_offset += 1
                i += 1

                if enable_general_comments:
                    general_comments.append(
                        self.create_general_comment(
                            review,
                            issue_opened=True,
                            issue_status=BaseComment.OPEN,
                            timestamp=timestamp))
                    issue_counts['open'] += 1
                    issue_counts['total'] += 1

                if enable_diffs and enable_diff_comments:
                    diff_comments.append(
                        self.create_diff_comment(
                            review,
                            filediffs[i],
                            issue_opened=True,
                            issue_status=BaseComment.RESOLVED,
                            timestamp=timestamp))
                    issue_counts['resolved'] += 1
                    issue_counts['total'] += 1

                if enable_file_attachments and enable_file_attachment_comments:
                    file_attachment_comments.append(
                        self.create_file_attachment_comment(
                            review,
                            file_attachments[i],
                            issue_opened=True,
                            issue_status=BaseComment.DROPPED,
                            timestamp=timestamp))
                    issue_counts['dropped'] += 1
                    issue_counts['total'] += 1

                if enable_screenshots and enable_screenshot_comments:
                    screenshot_comments.append(
                        self.create_screenshot_comment(
                            review,
                            screenshots[i],
                            issue_opened=False,
                            timestamp=timestamp))

        self.all_comments = [
            *general_comments,
            *screenshot_comments,
            *file_attachment_comments,
            *diff_comments,
        ]
        self.diff_comments = diff_comments
        self.general_comments = general_comments
        self.file_attachment_comments = file_attachment_comments
        self.screenshot_comments = screenshot_comments
        self.issue_counts = issue_counts
        self.reviews = reviews

        # Create some change descriptions.
        #
        # Only the second will have a user assigned (simulating older
        # change descriptions).
        changedescs: list[ChangeDescription] = []

        if enable_changedescs:
            for i in range(2):
                changedescs.append(review_request.changedescs.create(
                    timestamp=now + timedelta(days=days_offset + i),
                    public=True))

            days_offset += 2

            for i in range(3):
                changedescs.append(review_request.changedescs.create(
                    timestamp=now + timedelta(days=days_offset + i),
                    public=True,
                    user=user))

            days_offset += 3

        self.changedescs = changedescs

        # Create some status updates.
        status_updates: list[StatusUpdate] = []

        if enable_status_updates:
            status_updates += [
                self.create_status_update(review_request),
                self.create_status_update(review_request),
            ]

        self.status_updates = status_updates

        # Clear the review request's caches so we'll re-fetch the draft.
        review_request.clear_local_caches()



"""Unit tests for reviewboard.datagrids.columns.RepositoryColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import RepositoryColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class RepositoryColumnTests(BaseColumnTestCase):
    """Testing reviewboard.datagrids.columns.RepositoryColumn.

    Version Added:
        8.0
    """

    column = RepositoryColumn()

    def test_render_data_with_no_repository(self) -> None:
        """Testing RepositoryColumn.render_data with no repository"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_repository(self) -> None:
        """Testing RepositoryColumn.render_data with repository"""
        repository = self.create_repository(tool_name='Test',
                                            name='Test <Repo>')
        review_request = self.create_review_request(
            repository=repository,
            publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, 'Test &lt;Repo&gt;')

    def test_to_json_with_no_repository(self) -> None:
        """Testing RepositoryColumn.to_json with no repository"""
        review_request = self.create_review_request(publish=True)

        self.assertIsNone(
            self.column.to_json(self.stateful_column, review_request))

    def test_to_json_with_repository(self) -> None:
        """Testing RepositoryColumn.to_json with repository"""
        repository = self.create_repository(name='My Repo', tool_name='Test')
        review_request = self.create_review_request(
            repository=repository,
            publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            repository)



"""Review Board API errors.

This builds upon the :py:mod:`djblets.webapi.errors` module, which contains
API errors provided by Djblets.
"""

from djblets.webapi.errors import WebAPIError


#: An error indicating a diff revision was missing in the request.
UNSPECIFIED_DIFF_REVISION = WebAPIError(
    code=200,
    msg='Diff revision not specified.',
    error_type='diff-revision-missing',
    http_status=400)  # 400 Bad Request


#: An error indicating a provided diff revision was invalid.
INVALID_DIFF_REVISION = WebAPIError(
    code=201,
    msg='Invalid diff revision.',
    error_type='diff-revision-invalid',
    http_status=404)  # 404 Not Found


#: An error indicating a provided action was invalid.
#:
#: This is not used by Review Board and is considered deprecated.
#:
#: Deprecated:
#:     6.0
INVALID_ACTION = WebAPIError(
    code=202,
    msg='Invalid action specified.',
    error_type='action-invalid',
    http_status=400)  # 400 Bad Request


#: An error indicating a commit ID could not be found in the repository.
INVALID_CHANGE_NUMBER = WebAPIError(
    code=203,
    msg='The commit ID specified could not be found.',
    error_type='repository-commit-id-invalid',
    http_status=404)  # 404 Not Found


#: An error indicating a commit ID is in use by another review request.
#:
#: For historical reasons, this is only used when creating review requests
#: or drafts, and applies to ``commit_id`` or ``changenum`` values.
#:
#: See :py:data:`COMMIT_ID_ALREADY_EXISTS`.
CHANGE_NUMBER_IN_USE = WebAPIError(
    code=204,
    msg='The commit ID specified has already been used.',
    error_type='review-request-commit-id-conflict',
    http_status=409)  # 409 Conflict


#: An error indicating a repository could not be found at a path.
MISSING_REPOSITORY = WebAPIError(
    code=205,
    msg='There was no repository found at the specified path.',
    error_type='repository-missing',
    http_status=400)  # 400 Bad Request


#: An error indicating a provided repository path/ID is invalid.
INVALID_REPOSITORY = WebAPIError(
    code=206,
    msg=(
        'The repository path specified is not in the list of known '
        'repositories.'
    ),
    error_type='repository-invalid',
    http_status=400)  # 400 Bad Request


#: An error indicating a file could not be found in a repository.
REPO_FILE_NOT_FOUND = WebAPIError(
    code=207,
    msg='The file was not found in the repository.',
    error_type='repository-file-not-found',
    http_status=400)  # 400 Bad Request


#: An error indicating a provided user does not exist.
#:
#: This is used when adding users to repositories, groups, or review requests
#: as reviewers.
INVALID_USER = WebAPIError(
    code=208,
    msg='User does not exist.',
    error_type='user-invalid',
    http_status=400)  # 400 Bad Request


#: An error indicating a repository request is not supported.
#:
#: This is used when a repository does not support fetching additional
#: information, such as repository metadata, commits, or branches.
REPO_NOT_IMPLEMENTED = WebAPIError(
    code=209,
    msg='The specified repository is not able to perform this action.',
    error_type='repository-request-not-supported',
    http_status=501)  # 501 Not Implemented


#: An error indicating failure fetching information from a repository.
#:
#: This is used when there's a communication error fetching state or
#: performing actions on a repository.
REPO_INFO_ERROR = WebAPIError(
    code=210,
    msg=(
        'There was an error communicating with this repository.'
    ),
    error_type='repository-communication-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating a publish failed due to a lack of changes.
NOTHING_TO_PUBLISH = WebAPIError(
    code=211,
    msg='You attempted to publish a review request without any modifications.',
    error_type='review-request-not-modified',
    http_status=400)  # 400 Bad Request


#: An error indicating a commit ID represents an empty changeset.
EMPTY_CHANGESET = WebAPIError(
    code=212,
    msg='The commit ID specified represents an empty changeset.',
    error_type='repository-commit-id-changeset-empty',
    http_status=400)  # 400 Bad Request


#: An error indicating a problem storing server-side configuration state.
#:
#: This is used when failing to store settings, certificate approvals,
#: credentials, or other data on a Review Board server.
SERVER_CONFIG_ERROR = WebAPIError(
    code=213,
    msg='There was an error storing configuration on the server.',
    error_type='server-config-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating a stored SSH host key does not match the server.
BAD_HOST_KEY = WebAPIError(
    code=214,
    msg='The SSH key on the host does not match the stored key.',
    error_type='ssh-host-key-mismatch',
    http_status=403)  # 403 Forbidden


#: An error indicating a SSH host key from a server is not verified.
UNVERIFIED_HOST_KEY = WebAPIError(
    code=215,
    msg='The SSH key on the host is unverified.',
    error_type='ssh-host-key-not-verified',
    http_status=403)  # 403 Forbidden


#: An error indicating a SSL certificate from a server is not verified.
UNVERIFIED_HOST_CERT = WebAPIError(
    code=216,
    msg='The HTTPS certificate on the host is unverified.',
    error_type='ssl-certificate-not-verified',
    http_status=403)  # 403 Forbidden


#: An error indicating a public SSH key is not available for the server.
MISSING_USER_KEY = WebAPIError(
    code=217,
    msg=(
        'A public SSH key was requested, but no SSH key was available to send.'
    ),
    error_type='ssh-public-key-missing',
    http_status=403)  # 403 Forbidden


#: An error indicating an authentication error with a repository.
REPO_AUTHENTICATION_ERROR = WebAPIError(
    code=218,
    msg=(
        'Unable to authenticate with the repository using the provided '
        'credentials.'
    ),
    error_type='repository-auth-error',
    http_status=403)  # 403 Forbidden


#: An error indicating a provided diff is empty.
DIFF_EMPTY = WebAPIError(
    code=219,
    msg='The specified diff file is empty.',
    error_type='diff-empty',
    http_status=400)  # 400 Bad Request


#: An error indicating a provided diff exceeds configured size limits.
DIFF_TOO_BIG = WebAPIError(
    code=220,
    msg='The specified diff file is too large.',
    error_type='diff-too-large',
    http_status=400)  # 400 Bad Request


#: An error indicating a failure retrieving a file from a repository
#:
#: This is used when a file should be present but unexpectedly could not be
#: retrieved.
FILE_RETRIEVAL_ERROR = WebAPIError(
    code=221,
    msg='There was an error fetching a source file.',
    error_type='repository-file-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating an authenticatiion error with a hosting service.
HOSTINGSVC_AUTH_ERROR = WebAPIError(
    code=222,
    msg='There was an error authorizing with a service.',
    error_type='hosting-service-auth-error',
    http_status=403)  # 403 Forbidden


#: An error indicating a review group already exists.
GROUP_ALREADY_EXISTS = WebAPIError(
    code=223,
    msg='A group with this name already exists.',
    error_type='review-group-conflict',
    http_status=409)  # 409 Conflict


#: An error indicating a provided diff could not be parsed.
DIFF_PARSE_ERROR = WebAPIError(
    code=224,
    msg='The specified diff file could not be parsed.',
    error_type='diff-parse-error',
    http_status=400)  # 400 Bad Request


#: An error indicating a review request could not be published.
#:
#: This is used if there's an error publishing the review request due to
#: missing state, a validation error, or an extension-provided failure.
PUBLISH_ERROR = WebAPIError(
    code=225,
    msg='An error occurred during publishing.',
    error_type='review-request-publish-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating a user could not be queried from an auth backend.
USER_QUERY_ERROR = WebAPIError(
    code=226,
    msg='An error occurred querying the user list.',
    error_type='user-query-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating a commit ID is in use by another review request.
#:
#: For historical reasons, this is only used when updating review requests
#: or drafts.
#:
#: See :py:data:`CHANGE_NUMBER_IN_USE`.
COMMIT_ID_ALREADY_EXISTS = WebAPIError(
    code=227,
    msg=(
        'Review request with this commit ID already exists.',
    ),
    error_type='review-request-commit-id-conflict',
    http_status=409)  # 409 Conflict


#: An error indicating an API token could not be generated.
TOKEN_GENERATION_FAILED = WebAPIError(
    code=228,
    msg='There was an error generating the API token. Please try again.',
    error_type='api-token-generation-error',
    http_status=500)  # 500 Internal Server Error.


#: An error indicating a repository already exists.
#:
#: This is used when trying to create a repository that conflicts with
#: an existing repository.
REPOSITORY_ALREADY_EXISTS = WebAPIError(
    code=229,
    msg='A repository with this name already exists.',
    error_type='repository-conflict',
    http_status=409)  # 409 Conflict


#: An error indicating a failure closing a review request.
CLOSE_ERROR = WebAPIError(
    code=230,
    msg='An error occurred while closing the review request.',
    error_type='review-request-close-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating a failure reopening a review request.
REOPEN_ERROR = WebAPIError(
    code=231,
    msg='An error occurred while reopening the review request.',
    error_type='review-request-reopen-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating a failure revoking a Ship It!
REVOKE_SHIP_IT_ERROR = WebAPIError(
    code=232,
    msg='An error occurred while revoking the Ship It for a review.',
    error_type='review-request-revoke-ship-it-error',
    http_status=500)  # 500 Internal Server Error


#: An error indicating the site is in read-only mode.
#:
#: This is used when attempting to perform operations that may write to the
#: server without permission while in Read-Only Mode,
READ_ONLY_ERROR = WebAPIError(
    code=233,
    msg='The site is currently in read-only mode.',
    error_type='server-read-only',
    http_status=503)  # 503 Service Unavailable Error


#: An error indicating a review group does not exist.
INVALID_GROUP = WebAPIError(
    code=234,
    msg='Group does not exist.',
    error_type='review-group-invalid',
    http_status=400)  # 400 Bad Request
